(* ********************************************************

*********************************************************** *)

val TNOT = Define
`
TNOT P t =
    ~(P t)
`;

 
val EVENTUALLY = Define
`
EVENTUALLY P t1 =
    ?t2. t1 <= t2 /\  P t2
`;

val TAND = Define
`
TAND P1 P2 t =
    P1 t /\ P2 t
`;

set_fixity "TAND" (Infixr 450);



val TEQ = Define
`
TEQ f1 f2 t =
    (f1 t = f2 t)
`;

set_fixity "TEQ" (Infixr 450);

val TCON = Define
`
TCON c t =  c
`;


val EVENTUALLY_LEQ = store_thm("EVENTUALLY_LEQ",

(--`
EVENTUALLY P (t2+1) /\
 t <= t2
==>
(EVENTUALLY P (t+1))
`--),

RW_TAC arith_ss [EVENTUALLY] THEN
EXISTS_TAC (--`t2':num`--) THEN
RW_TAC arith_ss [] THEN
RES_TAC THEN
ZAP_TAC arith_ss []
);



val EVENTUALLY_TANDP_LEMMA = store_thm("EVENTUALLY_TANDP_LEMMA",
(--`
(!t. ~(P t) /\ Q t ==> Q (t + 1)) /\
 P t2 

==>
(!n. t2 >= n ==> Q (t2 - n) ==>  ?t3. t2 - n <= t3 /\ P t3 /\ Q t3)
   
`--),

STRIP_TAC THEN
Induct THENL
[
RW_TAC arith_ss [] THEN
EXISTS_TAC (--`t2:num`--) THEN
RW_TAC arith_ss []
,
RW_TAC arith_ss [] THEN
ASM_CASES_TAC (--`P (t2 - SUC n):bool`--) THENL
 [
   EXISTS_TAC (--`t2 - SUC n`--) THEN
   RW_TAC arith_ss []
 ,
   RES_TAC THEN
   IMP_RES_TAC 
      (DECIDE `(t2 >= SUC n) ==> (Q((t2 - SUC n) + 1) ==> Q( t2 - n))`) THEN
   IMP_RES_TAC (DECIDE `(t2 >= SUC n) ==> (t2 >= n)`) THEN
   RES_TAC THEN
   EXISTS_TAC (--`t3:num`--) THEN
   RW_TAC arith_ss []
 ]
]
);




(*val EVENTUALLY_TANDP = store_thm("EVENTUALLY_TANDP",
(--`
EVENTUALLY P t1 /\
Q t1 /\
(!t. (TNOT P) t /\ Q t ==> Q (t+1))
==>
(EVENTUALLY (P TAND Q) t1 = T)
`--),

RW_TAC arith_ss [EVENTUALLY,TAND,TNOT] THEN
IMP_RES_TAC EVENTUALLY_TANDP_LEMMA THEN
POP_ASSUM (fn th => ALL_TAC) THEN
POP_ASSUM (fn th => ALL_TAC) THEN
POP_ASSUM (fn th => ASSUME_TAC (SPEC (--`t2 - t1`--) th)) THEN
POP_ASSUM (fn th => ASSUME_TAC 
                         (REWRITE_RULE[(DECIDE `(t2 >= t2-t1)`)] th)) THEN
IMP_RES_TAC (DECIDE `t1 <= t2 ==> (t2 - t2 - t1 = t1)`) THEN
POP_ASSUM (fn th1 => (POP_ASSUM (fn th2 => 
              IMP_RES_TAC (REWRITE_RULE[th1] th2)))) THEN
EXISTS_TAC (--`t3:num`--) THEN
ASM_REWRITE_TAC[]
);

*)



 val EVENTUALLY_TANDP = store_thm("EVENTUALLY_TANDP",
(--`
EVENTUALLY P t1 /\
Q t1 /\
(!t. (TNOT P) t /\ Q t ==> Q (t+1))
==>
(EVENTUALLY (P TAND Q) t1 = T)
`--),

RW_TAC arith_ss [EVENTUALLY,TAND,TNOT] THEN
IMP_RES_TAC EVENTUALLY_TANDP_LEMMA THEN
POP_ASSUM (fn th => ALL_TAC) THEN
POP_ASSUM (fn th => ALL_TAC) THEN
POP_ASSUM (fn th => ASSUME_TAC (SPEC (--`t2 - t1`--) th)) THEN
POP_ASSUM (fn th => ASSUME_TAC 
                         (REWRITE_RULE[(DECIDE `(t2 >= t2-t1)`)] th)) THEN
IMP_RES_TAC (DECIDE `t1 <= t2 ==> (t2 - (t2 - t1) = t1)`) THEN
 ASSUM_LIST (fn asml => 
                            MP_TAC(REWRITE_RULE[] (el 2 asml))) THEN
 POP_ASSUM (fn t => PURE_ONCE_REWRITE_TAC[t]) THEN
REPEAT STRIP_TAC THEN
RES_TAC THEN
EXISTS_TAC (--`t3:num`--) THEN
ASM_REWRITE_TAC[]
);

(* The Specification of a User of the Chocolate Machine *)
(* ------------------------------------------------------------------------ *)
(* User States*)

val ustate_type = Type 
`: (num -> bool)
 # (num -> bool)
 # (num -> bool)
 # (num -> bool) 
 # (num -> num)
 # (num -> num)
 # (num -> num)

`;

val UserHasChoc = Define
`
UserHasChoc (ustate:^(ty_antiq ustate_type)) =  FST  ustate
`;

val UserHasChange = Define
`
UserHasChange (ustate:^(ty_antiq ustate_type)) = FST (SND ustate)
`;

val UserHasCoin = Define
`
UserHasCoin (ustate:^(ty_antiq ustate_type)) = FST (SND (SND ustate))
`;

val UserFinished = Define
`
UserFinished (ustate:^(ty_antiq ustate_type)) = FST  (SND (SND (SND ustate)))
`;


val CountChoc = Define
`
CountChoc (ustate:^(ty_antiq ustate_type)) = 
   FST (SND (SND (SND (SND ustate))))
`;

val CountChange = Define
`
CountChange (ustate:^(ty_antiq ustate_type)) = 
   FST (SND (SND (SND (SND (SND ustate)))))
`;

val CountCoin = Define
`
CountCoin (ustate:^(ty_antiq ustate_type)) = 
   (SND (SND (SND (SND (SND (SND ustate))))))
`;


(* ------------------------------------------------------------------------ *)



(* Axioms that are always true
     - not up to the whim of the user at a particular time
*)


(* We assume that when the machine gives out chocolate and change,
 it does so with a noisy clunk in a way that reasonable users 
 would not fail to take it - This may not be true of all machines *)


(* ****************************************************** *)



val HAS_POSSESSION = Define
`
HAS_POSSESSION  haspossession takepossession givepossession
                  valueofpossession countpossession
                  (ustate:'u) (opstate:(num->Mdg_Basic) list) (ipstate:(num->Mdg_Basic) list) =

      (!t. haspossession ustate t = (countpossession ustate t > 0)) /\
      (!t. (((givepossession ipstate t) = (BOOL T)) /\ ((takepossession opstate t) = (BOOL F))) =
             ((countpossession ustate t > 0) /\
              (countpossession ustate (t+1) = countpossession ustate t - 1 ))) /\
      (!t. (((takepossession opstate t) = (BOOL T)) /\ ((givepossession ipstate t) = (BOOL F))) =
             (countpossession ustate (t+1) = countpossession ustate t + 1 )) /\
      (!t.
          ((((givepossession ipstate t) = (BOOL F))) /\ ((takepossession opstate t) = (BOOL F))) \/
           (((givepossession ipstate t) = (BOOL T)) /\ ((takepossession opstate t) = (BOOL T))) =
                (countpossession ustate (t+1) = countpossession ustate t))

`handle e=>Raise e;


(* ****************************************************** *)


val pstate_type = Type 
`:('u -> num -> bool) # ((num->Mdg_Basic) list -> num -> Mdg_Basic) # ((num->Mdg_Basic) list -> num -> Mdg_Basic) # num #
  ('u -> num -> num)`;


val HasPossession = Define
`
HasPossession (pstate:^(ty_antiq pstate_type)) =  FST  pstate
`;

val TakePossession = Define
`
TakePossession (pstate:^(ty_antiq pstate_type)) = FST (SND pstate)
`;

val GivePossession = Define
`
GivePossession (pstate:^(ty_antiq pstate_type)) = FST (SND (SND pstate))
`;

val ValueOfPossession = Define
`
ValueOfPossession (pstate:^(ty_antiq pstate_type)) = FST (SND (SND (SND pstate)))
`;

val CountPossession = Define
`
CountPossession (pstate:^(ty_antiq pstate_type)) = SND (SND (SND (SND pstate)))
`;



(* ****************************************************** *)

val POSSESSIONS = Define
`
(POSSESSIONS [] (ustate:'u) (opstate:(num->Mdg_Basic) list) (ipstate:(num->Mdg_Basic) list) = T) /\
(POSSESSIONS (CONS (p:^(ty_antiq pstate_type)) possessions) ustate opstate ipstate=
    ((POSSESSIONS possessions ustate opstate ipstate) /\
     (HAS_POSSESSION (HasPossession p) (TakePossession p)
                     (GivePossession p) (ValueOfPossession p)
                     (CountPossession p)
                     ustate opstate ipstate)))
`;




val POSSESSIONS_VAL = Define
`
(POSSESSIONS_VAL []  (ustate:'u) t = 0) /\
(POSSESSIONS_VAL (CONS (p:^(ty_antiq pstate_type)) ps)  (ustate:'u) t =
    ((POSSESSIONS_VAL ps (ustate:'u) t) +
     ((CountPossession p ustate t) * (ValueOfPossession p))))
`;

(* ****************************************************** *)

(* ****************************************************** *)




 val LIGHT = Define
`
LIGHT light (action :(num -> Mdg_Basic) list -> num -> Mdg_Basic) (opstate:(num->Mdg_Basic) list) (ipstate:(num->Mdg_Basic) list) t =
    (light opstate t =  (BOOL T))  /\
    EVENTUALLY  (\t. (action ipstate t = BOOL T)) t
`;

val LIGHTS = Define
`
(LIGHTS [] (opstate:(num->Mdg_Basic) list)  (ipstate:(num->Mdg_Basic) list) t = F) /\
(LIGHTS (CONS l lights_actions) opstate ipstate t =
    ((LIGHTS lights_actions opstate ipstate t) \/
     (LIGHT (FST l) (SND l) opstate ipstate t)))
`;



(*---------------------------------------------------------------------------*)

(* Once finished a user never ceases to have finished *)
val GENERAL_USER_UNIVERSAL = Define
`
GENERAL_USER_UNIVERSAL possessions finished (ustate:'u) (opstate:(num->Mdg_Basic) list)  (ipstate:(num->Mdg_Basic) list) =
    (!t. finished ustate t ==> finished ustate (t+1)) /\
    (POSSESSIONS possessions ustate opstate ipstate)
`;

(* One possibility is that the user finishes when the goal is reached
   but it is not guaranteed *)
val COMPLETION = Define
`
COMPLETION finished goalachieved (ustate:'u) (t:num) =
    (goalachieved ustate t = T) /\
    EVENTUALLY (finished ustate) t
`;

val GENERAL_USER_CHOICE = Define
`
GENERAL_USER_CHOICE lights_actions finished goalachieved (opstate:(num->Mdg_Basic) list) (ipstate:(num->Mdg_Basic) list) (ustate:'u) t =
    LIGHTS lights_actions opstate ipstate t \/
    (COMPLETION finished goalachieved ustate t)
`;

val USER = Define
` 
USER lights_actions possessions
     finished goalachieved invariant (ustate:'u) (opstate:(num->Mdg_Basic) list) (ipstate:(num->Mdg_Basic) list) =
   (GENERAL_USER_UNIVERSAL possessions finished ustate opstate ipstate) /\
   (!t. 
       (((invariant ustate t = T) /\ 
         (goalachieved ustate t = T)) =>
               EVENTUALLY (finished ustate) t |
               GENERAL_USER_CHOICE lights_actions finished goalachieved
                                    opstate ipstate  ustate t))
`;
(* ********************************************************************* *)
val VALUE_INVARIANT = Define
` 
VALUE_INVARIANT  possessions (ustate:'u) t =
   (POSSESSIONS_VAL possessions ustate t >=
                             POSSESSIONS_VAL possessions ustate 0)
`;


val CHOC_POSSESSIONS = Define
` 
CHOC_POSSESSIONS haschoc givechoc countchoc
                 haschange givechange countchange
                 hascoin insertcoin countcoin =

       [(* HAS,       TAKE,    GIVE UP,     VALUE,	COUNT *)

        (haschoc,   givechoc,   (\m t.(BOOL  F)),    50,	countchoc);
        (haschange, givechange,  (\(m:(num->Mdg_Basic) list) t. (BOOL F)),    50,	countchange);
        (hascoin, (\(m:(num->Mdg_Basic) list) t. (BOOL F)),    insertcoin,   100,	countcoin)
       ]
`;

val CHOC_MACHINE_USER = Define
` 
CHOC_MACHINE_USER (ustate:^(ty_antiq ustate_type)) (opstate:(num->Mdg_Basic) list)(ipstate:(num->Mdg_Basic) list) =

      USER 
       [(CoinLight,InsertCoin);
        (ChocLight,PushChoc);
        (ChangeLight,PushChange)
       ]
       (CHOC_POSSESSIONS UserHasChoc GiveChoc CountChoc
                         UserHasChange GiveChange CountChange
                         UserHasCoin InsertCoin	CountCoin)
       UserFinished
       UserHasChoc 
       (VALUE_INVARIANT (CHOC_POSSESSIONS UserHasChoc GiveChoc CountChoc
                         UserHasChange GiveChange CountChange
                         UserHasCoin InsertCoin	CountCoin))
       ustate opstate ipstate
`;

(*-----------------------------------------------------------*)

val CHOC_MACHINE_USABLE = Define
`
CHOC_MACHINE_USABLE (ustate:^(ty_antiq ustate_type)) (opstate:(num->Mdg_Basic) list)(ipstate:(num->Mdg_Basic) list)  =
    !t. ~(UserHasChoc ustate t) /\ (* ie Not achived goal yet *)
        ~(UserHasChange ustate t) /\
        (UserHasCoin ustate t) /\
        (VALUE_INVARIANT (CHOC_POSSESSIONS UserHasChoc GiveChoc CountChoc
                         UserHasChange GiveChange CountChange
                         UserHasCoin InsertCoin	CountCoin) ustate t)  /\ 
        ((CoinLight opstate t)= BOOL T) ==>
               ?t1.
                      (UserHasChoc ustate t1) /\
                      (UserHasChange ustate t1)
`;

